# YAML helpers from http://stackoverflow.com/questions/8640959

import yaml
from collections import OrderedDict

class quoted(str): pass

def quoted_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='"')

class literal(str): pass

def literal_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')

def ordered_dict_presenter(dumper, data):
    return dumper.represent_dict(data.items())

def add_representers():
    yaml.add_representer(quoted, quoted_presenter)
    yaml.add_representer(literal, literal_presenter)
    yaml.add_representer(OrderedDict, ordered_dict_presenter)


import autonetkit.log as log

class PlatformCompiler(object):
    """Base Platform Compiler"""
# and set properties in nidb._graph.graph
    def __init__(self, nidb, anm, host):
        self.nidb = nidb
        self.anm = anm
        self.host = host

    @property
    def timestamp(self):
        return self.nidb.timestamp

    def compile(self):
        # TODO: make this abstract
        pass

    def copy_across_ip_addresses(self):
        #TODO: try/except and raise SystemError as fatal error if cant copy
        from autonetkit.ank import sn_preflen_to_network

        #TODO:  check if this will clobber with platform?
        for node in self.nidb.l3devices(host=self.host):
            phy_node = self.anm['phy'].node(node)

            node.add_stanza("ip")
            node.ip.use_ipv4 = phy_node.use_ipv4 or False
            node.ip.use_ipv6 = phy_node.use_ipv6 or False

            for interface in node.interfaces:
                phy_int = phy_node.interface(interface)
                if phy_node.use_ipv4:
                    ipv4_int = phy_int['ipv4']
                    if node.is_server() and interface.is_loopback:
                        continue
                    if interface.is_physical and not interface.is_bound:
                        continue
                    # interface is connected
                    interface.use_ipv4 = True
                    interface.ipv4_address = ipv4_int.ip_address
                    interface.ipv4_subnet = ipv4_int.subnet
                    interface.ipv4_cidr = sn_preflen_to_network(interface.ipv4_address,
                            interface.ipv4_subnet.prefixlen)
                if phy_node.use_ipv6:
                    ipv6_int = phy_int['ipv6']
                    if node.is_server() and interface.is_loopback:
                        continue
                    if interface.is_physical and not interface.is_bound:
                        continue
                    interface.use_ipv6 = True
                    try:
                        #TODO: copy ip address as well
                        interface.ipv6_subnet = ipv6_int.subnet
                        interface.ipv6_address = sn_preflen_to_network(ipv6_int.ip_address,
                            interface.ipv6_subnet.prefixlen)
                    except AttributeError:
                        log.warning("Unable to copy IPv6 subnet for %s" % interface)




"""Module to build overlay graphs for network design"""
import itertools

import autonetkit
import autonetkit.ank as ank_utils
import autonetkit.ank_messaging as ank_messaging
import autonetkit.anm
import autonetkit.config
import autonetkit.exception
import autonetkit.load.graphml as graphml
import autonetkit.log as log
import networkx as nx
from autonetkit.exception import AutoNetkitException

SETTINGS = autonetkit.config.settings

#TODO: revisit phy_neighbors for eg ASN and use l3_conn instead

#TODO: note that build network now assumes input graph has interface mappings on nodes/edges

__all__ = ['build']
from autonetkit.ank_utils import call_log

@call_log
def load(input_graph_string):
    #TODO: look at XML header for file type
    try:
        input_graph = graphml.load_graphml(input_graph_string)
    except autonetkit.exception.AnkIncorrectFileFormat:
# try a different reader
        try:
            from autonetkit_cisco import load as cisco_load
        except ImportError, e:
            log.debug("Unable to load autonetkit_cisco %s" % e)
            return  # module not present (development module)
        else:
            input_graph = cisco_load.load(input_graph_string)
            # add local deployment host
            SETTINGS['General']['deploy'] = True
            SETTINGS['Deploy Hosts']['internal'] = {
                        'VIRL': {
                        'deploy': True,
                        },
                        }

    return input_graph

@call_log
def grid_2d(dim):
    """Creates a 2d grid of dimension dim"""
    graph = nx.grid_2d_graph(dim, dim)

    for node in graph:
        graph.node[node]['asn'] = 1
        graph.node[node]['x'] = node[0] * 150
        graph.node[node]['y'] = node[1] * 150
        graph.node[node]['device_type'] = 'router'
        graph.node[node]['platform'] = 'cisco'
        graph.node[node]['syntax'] = 'ios_xr'
        graph.node[node]['host'] = 'internal'
        graph.node[node]['ibgp_role'] = "Peer"

    mapping = {node: "%s_%s" % (node[0], node[1]) for node in graph}
    # Networkx wipes data if remap with same labels
    nx.relabel_nodes(graph, mapping, copy=False)
    for index, (src, dst) in enumerate(graph.edges()):
        graph[src][dst]['type'] = "physical"
        # add global index for sorting

    SETTINGS['General']['deploy'] = True
    SETTINGS['Deploy Hosts']['internal'] = {
        'cisco': {
        'deploy': True,
        },
    }

    return graph

@call_log
def initialise(input_graph):
    """Initialises the input graph with from a NetworkX graph"""
    anm = autonetkit.anm.AbstractNetworkModel()

    input_undirected = nx.Graph(input_graph)
    g_in = anm.add_overlay("input", graph=input_undirected)

# set defaults
    if not g_in.data.specified_int_names:
        # if not specified then automatically assign interface names
        g_in.data.specified_int_names = False

    #import autonetkit.plugins.graph_product as graph_product
    #graph_product.expand(g_in)  # apply graph products if relevant

    expand_fqdn = False
    # TODO: make this set from config and also in the input file
    if expand_fqdn and len(ank_utils.unique_attr(g_in, "asn")) > 1:
        # Multiple ASNs set, use label format device.asn
        anm.set_node_label(".", ['label', 'asn'])

    g_in.update(g_in.routers(platform="junosphere"), syntax="junos")
    g_in.update(g_in.routers(platform="dynagen"), syntax="ios")
    g_in.update(g_in.routers(platform="netkit"), syntax="quagga")
    #TODO: is this used?
    g_in.update(g_in.servers(platform="netkit"), syntax="quagga")

    autonetkit.ank.set_node_default(g_in,  specified_int_names=None)

    g_graphics = anm.add_overlay("graphics")  # plotting data
    g_graphics.add_nodes_from(g_in, retain=['x', 'y', 'device_type',
        'label', 'device_subtype', 'asn'])

    if g_in.data.Creator == "VIRL":
        #TODO: move this to other module
        # Multiple ASNs set, use label format device.asn
        #anm.set_node_label(".", ['label_full'])
        pass

    return anm

@call_log
def check_server_asns(anm):
    """Checks that servers have appropriate ASN allocated.
    Warns and auto-corrects servers which are connected to routers of a difference AS.
    #TODO: provide manual over-ride for this auto-correct.
    """
    #TODO: Move to validate module?
    g_phy = anm['phy']

    for server in g_phy.servers():
        if server.device_subtype in ("SNAT", "FLAT"):
            continue # Don't warn on ASN for NAT elements
        l3_neighbors = list(server['l3_conn'].neighbors())
        l3_neighbor_asns = set(n.asn for n in l3_neighbors)
        if server.asn not in l3_neighbor_asns:
            neighs_with_asn = ["%s: AS %s" % (n, n.asn)
                for n in l3_neighbors] # tuples for warning message
            server.log.warning("Server does not belong to same ASN as neighbors %s" % (neighs_with_asn))

            if len(l3_neighbors) == 1:
                # single ASN of neighbor -> auto correct
                if server['input'].default_asn:
                    neigh_asn = l3_neighbor_asns.pop()
                    log.warning("Updating server %s AS from %s to %s" % (server, server.asn, neigh_asn))
                    server.asn = neigh_asn
                else:
                    log.info("Server %s ASN %s explictly set by user, not auto-correcting" %
                        (server, server.asn))


@call_log
def apply_design_rules(anm):
    """Applies appropriate design rules to ANM"""
    g_in = anm['input']

    build_phy(anm)
    g_phy = anm['phy']

    build_l3_connectivity(anm)
    check_server_asns(anm)

    from autonetkit.design.mpls import build_vrf
    build_vrf(anm) # need to do before to add loopbacks before ip allocations
    from autonetkit.design.ip import build_ip, build_ipv4, build_ipv6
    #TODO: replace this with layer2 overlay topology creation
    build_ip(anm) # ip infrastructure topology

#TODO: set defaults at the start, rather than inline, ie set g_in.data.address_family then use later

    address_family = g_in.data.address_family or "v4" # default is v4
#TODO: can remove the infrastructure now create g_ip seperately
    if address_family == "None":
        log.info("IP addressing disabled, disabling routing protocol configuration")
        anm['phy'].data.enable_routing = False

    if address_family == "None":
        log.info("IP addressing disabled, skipping IPv4")
        anm.add_overlay("ipv4") # create empty so rest of code follows through
        g_phy.update(g_phy, use_ipv4 = False)
    elif address_family in ("v4", "dual_stack"):
        build_ipv4(anm, infrastructure = True)
        g_phy.update(g_phy, use_ipv4 = True)
    elif address_family == "v6":
        # Allocate v4 loopbacks for router ids
        build_ipv4(anm, infrastructure = False)
        g_phy.update(g_phy, use_ipv4 = False)

    #TODO: Create a collision domain overlay for ip addressing - l2 overlay?
    if address_family == "None":
        log.info("IP addressing disabled, not allocating IPv6")
        anm.add_overlay("ipv6") # create empty so rest of code follows through
        g_phy.update(g_phy, use_ipv6 = False)
    elif address_family in ("v6", "dual_stack"):
        build_ipv6(anm)
        g_phy.update(g_phy, use_ipv6 = True)
    else:
        anm.add_overlay("ipv6") # placeholder for compiler logic

    default_igp = g_in.data.igp or "ospf"
    ank_utils.set_node_default(g_in,  igp=default_igp)

    ank_utils.copy_attr_from(g_in, g_phy, "include_csr")

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError, e:
        log.debug("Unable to load autonetkit_cisco %s" % e)
    else:
        cisco_build_network.pre_design(anm)

    from autonetkit.design.igp import build_ospf, build_eigrp, build_isis
    build_ospf(anm)
    build_eigrp(anm)
    build_isis(anm)

    from autonetkit.design.bgp import build_bgp
    build_bgp(anm)
    #autonetkit.update_http(anm)

    from autonetkit.design.mpls import mpls_te, mpls_oam
    mpls_te(anm)
    mpls_oam(anm)

# post-processing
    if anm['phy'].data.enable_routing:
        from autonetkit.design.mpls import mark_ebgp_vrf, build_ibgp_vpn_v4
        mark_ebgp_vrf(anm)
        build_ibgp_vpn_v4(anm) # build after bgp as is based on
    #autonetkit.update_http(anm)

    try:
        from autonetkit_cisco import build_network as cisco_build_network
    except ImportError, e:
        log.debug("Unable to load autonetkit_cisco %s" % e)
    else:
        cisco_build_network.post_design(anm)

    return anm


@call_log
def build(input_graph):
    """Main function to build network overlay topologies"""
    anm = None
    try:
        anm = initialise(input_graph)
        anm = apply_design_rules(anm)
        #print {str(node): {'x': node.x, 'y': node.y} for node in anm['input']}
        import autonetkit
        autonetkit.update_http(anm)
    except Exception, e:
        # Send the visualisation to help debugging
        import autonetkit
        try:
            autonetkit.update_http(anm)
        except Exception, e:
            # problem with vis -> could be coupled with original exception - raise original
            log.warning("Unable to visualise: %s" % e)
        raise # raise the original exception
    return anm


def remove_parallel_switch_links(anm):
    g_phy = anm['phy']
    subs = ank_utils.connected_subgraphs(g_phy, g_phy.switches())
    for component in subs:
        log.debug("Checking for multiple links to switch cluster %s" % str(sorted(component)))

        # Collect all links into this cluster
        external_edges = []
        for switch in component:
            for edge in switch.edges():
                if edge.dst not in component:
                    external_edges.append(edge)

        # Group by the node they link to
        from collections import defaultdict
        check_dict = defaultdict(list)
        for edge in external_edges:
            check_dict[edge.dst].append(edge)

        # Check to see if any nodes have more than one link into this aggregate
        for dst, edges in check_dict.items():
            if len(edges) > 1:
                edges_to_remove = sorted(edges)[1:] # remove all but first
                interfaces = ", ".join(sorted(str(edge.dst_int['phy']) for edge in edges))
                interfaces_to_disconnect = ", ".join(sorted(str(edge.dst_int['phy']) for edge in edges_to_remove))
                dst.log.warning("Multiple edges exist to same switch cluster: %s (%s). Removing edges from interfaces %s" % (
                    str(sorted(component)), interfaces, interfaces_to_disconnect))

                g_phy.remove_edges_from(edges_to_remove)

@call_log
def build_phy(anm):
    """Build physical overlay"""
    g_in = anm['input']
    g_phy = anm['phy']

    g_phy.data.enable_routing = g_in.data.enable_routing
    if g_phy.data.enable_routing is None:
        g_in.data.enable_routing = True # default if not set

    g_phy.add_nodes_from(g_in, retain=['label', 'update', 'device_type', 'asn',
        'specified_int_names',
        'device_subtype', 'platform', 'host', 'syntax'])
    if g_in.data.Creator == "Topology Zoo Toolset":
        ank_utils.copy_attr_from(g_in, g_phy, "Network")

    ank_utils.set_node_default(g_phy,  Network=None)
    g_phy.add_edges_from(g_in.edges(type="physical"))
    # TODO: make this automatic if adding to the physical graph?

    if g_in.data.Creator == "VIRL":
        g_phy.data.mgmt_interfaces_enabled = g_in.data.mgmt_interfaces_enabled
        #TODO: remove this code now allocated externally
        g_phy.data.mgmt_address_start = g_in.data.mgmt_address_start
        g_phy.data.mgmt_address_end = g_in.data.mgmt_address_end
        g_phy.data.mgmt_prefixlen = g_in.data.mgmt_prefixlen
        g_phy.data.mgmt_prefixlen = g_in.data.mgmt_prefixlen

        ank_utils.copy_attr_from(g_in, g_phy, "use_cdp")
        ank_utils.copy_attr_from(g_in, g_phy, "use_onepk")
        ank_utils.copy_attr_from(g_in, g_phy, "label_full")
        ank_utils.copy_attr_from(g_in, g_phy, "indices")
        ank_utils.copy_attr_from(g_in, g_phy, "dont_configure_static_routing")
        ank_utils.copy_attr_from(g_in, g_phy, "server_username")
        ank_utils.copy_attr_from(g_in, g_phy, "server_ssh_key")

    ank_utils.set_node_default(g_phy,  use_ipv4 = False, use_ipv6=False)

    g_phy.allocate_interfaces()

    for node in g_phy:
        for interface in node:
            specified_id = interface['input'].get("specified_id")
            if specified_id:
                interface.specified_id = specified_id # map across

    remove_parallel_switch_links(anm)

@call_log
def build_l3_connectivity(anm):
    """ creates l3_connectivity graph, which is switch nodes aggregated and exploded"""
    #TODO: use this as base for ospf, ebgp, ip, etc rather than exploding in each
    g_in = anm['input']
    g_l3conn = anm.add_overlay("l3_conn")
    g_l3conn.add_nodes_from(g_in, retain=['label', 'update', 'device_type', 'asn',
        'specified_int_names',
        'device_subtype', 'platform', 'host', 'syntax'])
    g_l3conn.add_nodes_from(g_in.switches(), retain=['asn'])
    g_l3conn.add_edges_from(g_in.edges())

    ank_utils.aggregate_nodes(g_l3conn, g_l3conn.switches())
    exploded_edges = ank_utils.explode_nodes(g_l3conn,
        g_l3conn.switches())
    for edge in exploded_edges:
        edge.multipoint = True
        edge.src_int.multipoint = True
        edge.dst_int.multipoint = True

@call_log
def build_conn(anm):
    """Build connectivity overlay"""
    g_in = anm['input']
    g_conn = anm.add_overlay("conn", directed=True)
    g_conn.add_nodes_from(g_in, retain=['label'])
    g_conn.add_edges_from(g_in.edges(type="physical"))

    return



#!/usr/bin/python
# -*- coding: utf-8 -*-
import itertools
import math
import os
import string
from collections import defaultdict
from cStringIO import StringIO

import autonetkit.config
import autonetkit.exception
import autonetkit.log as log
import networkx as nx

settings = autonetkit.config.settings


# TODO: fallback to python if cStringIO version is not available

def load_graphml(input_data):

    # TODO: allow default properties to be passed in as dicts

    try:
        graph = nx.read_graphml(input_data)
    except IOError, e:
        acceptable_errors = set([2, 36, 63])  # no such file or directory
                                              # input string too long for filename
                                              # input string too long for filename
        if e.errno in acceptable_errors:
            from xml.etree.cElementTree import ParseError

            # try as data string rather than filename string

            try:
                input_pseduo_fh = StringIO(input_data)  # load into filehandle to networkx
                graph = nx.read_graphml(input_pseduo_fh)
            except IOError:
                raise autonetkit.exception.AnkIncorrectFileFormat
            except IndexError:
                raise autonetkit.exception.AnkIncorrectFileFormat
            except ParseError:
                raise autonetkit.exception.AnkIncorrectFileFormat
            except ParseError:
                raise autonetkit.exception.AnkIncorrectFileFormat
        else:
            raise e

    if graph.is_multigraph():
        log.info('Input graph is multigraph. Converting to single-edge graph'
                 )
        if graph.is_directed():
            graph = nx.DiGraph(graph)
        else:
            graph = nx.Graph(graph)

    # TODO: need to support edge index keying for multi graphs

    graph.remove_edges_from(edge for edge in graph.selfloop_edges())

# TODO: if selfloops then log that are removing

    letters_single = (c for c in string.lowercase)  # a, b, c, ... z
    letters_double = ('%s%s' % (a, b) for (a, b) in
        itertools.product(string.lowercase, string.lowercase))  # aa, ab, ... zz
    letters = itertools.chain(letters_single, letters_double)  # a, b, c, .. z, aa, ab, ac, ... zz

# TODO: need to get set of current labels, and only return if not in this set

    # TODO: add cloud, host, etc
    # prefixes for unlabelled devices, ie router -> r_a

    label_prefixes = {'router': 'r', 'switch': 'sw', 'server': 'se'}

    current_labels = set(graph.node[node].get('label') for node in
                         graph.nodes_iter())
    unique_label = (letter for letter in letters if letter
        not in current_labels)

# TODO: make sure device label set

    ank_graph_defaults = settings['Graphml']['Graph Defaults']
    for (key, val) in ank_graph_defaults.items():
        if key not in graph.graph:
            graph.graph[key] = val

    # handle yEd exported booleans: if a boolean is set, then only the nodes marked true have the attribute. need to map the remainder to be false to allow ANK logic
    # for node in graph.nodes(data=True):
        # print node

    all_labels = dict((n, d.get('label')) for (n, d) in
                      graph.nodes(data=True))
    label_counts = defaultdict(list)
    for (node, label) in all_labels.items():
        label_counts[label].append(node)

    # set default name for blank labels to ensure unique

    try:
        blank_labels = [v for (k, v) in label_counts.items()
                        if not k].pop()  # strip outer list
    except IndexError:
        blank_labels = []  # no blank labels
    for (index, node) in enumerate(blank_labels):

        # TODO: log message that no label set, so setting default

        graph.node[node]['label'] = 'none___%s' % index

    duplicates = [(k, v) for (k, v) in label_counts.items() if k
                  and len(v) > 1]
    for (label, nodes) in duplicates:
        for node in nodes:

            # TODO: need to check they don't all have same ASN... if so then warn

            try:
                graph.node[node]['label'] = '%s_%s' \
                    % (graph.node[node]['label'], graph.node[node]['asn'
                       ])
            except KeyError:
                log.warning('Unable to set new label for duplicate node %s: %s'
                             % (node, graph.node[node].get('label')))

    boolean_attributes = set(k for (n, d) in graph.nodes(data=True)
                             for (k, v) in d.items() if isinstance(v,
                             bool))

    for node in graph:
        for attr in boolean_attributes:
            if attr not in graph.node[node]:
                graph.node[node][attr] = False

    boolean_attributes = set(k for (n1, d1) in graph.edge.items()
                             for (n2, d2) in d1.items() for (k, v) in
                             d2.items() if isinstance(v, bool))
    for (n1, d1) in graph.edge.items():
        for (n2, d2) in d1.items():
            for attr in boolean_attributes:
                if attr not in graph.edge[n1][n2]:
                    graph.edge[n1][n2][attr] = False

# TODO: store these in config file

    ank_node_defaults = settings['Graphml']['Node Defaults']
    node_defaults = graph.graph['node_default']  # update with defaults from graphml
    for (key, val) in node_defaults.items():
        if val == 'False':
            node_defaults[key] = False

    # TODO: do a dict update before applying so only need to iterate nodes once

    for (key, val) in ank_node_defaults.items():
        if key not in node_defaults or node_defaults[key] == 'None':
            node_defaults[key] = val

    for node in graph:
        for (key, val) in node_defaults.items():
            if key not in graph.node[node]:
                graph.node[node][key] = val

    # set address family

    graph.graph['address_family'] = 'v4'
    graph.graph['enable_routing'] = True

    # map lat/lon from zoo to crude x/y approximation

    if graph.graph.get('Creator') == 'Topology Zoo Toolset':
        all_lat = [graph.node[n].get('Latitude') for n in graph
                   if graph.node[n].get('Latitude')]
        all_lon = [graph.node[n].get('Longitude') for n in graph
                   if graph.node[n].get('Longitude')]

        lat_min = min(all_lat)
        lon_min = min(all_lon)
        lat_max = max(all_lat)
        lon_max = max(all_lon)
        lat_mean = (lat_max - lat_min) / 2
        lon_mean = (lon_max - lon_min) / 2
        lat_scale = 500 / (lat_max - lat_min)
        lon_scale = 500 / (lon_max - lon_min)
        for node in graph:
            lat = graph.node[node].get('Latitude') or lat_mean  # set default to be mean of min/max
            lon = graph.node[node].get('Longitude') or lon_mean  # set default to be mean of min/max
            graph.node[node]['y'] = -1 * lat * lat_scale
            graph.node[node]['x'] = lon * lon_scale

    if not (any(graph.node[n].get('x') for n in graph)
            and any(graph.node[n].get('y') for n in graph)):

# No x, y set, layout in a grid

        grid_length = int(math.ceil(math.sqrt(len(graph))))
        co_ords = [(x * 100, y * 100) for y in range(grid_length)
                   for x in range(grid_length)]

        # (0,0), (100, 0), (200, 0), (0, 100), (100, 100) ....

        for node in sorted(graph):
            (x, y) = co_ords.pop(0)
            graph.node[node]['x'] = x
            graph.node[node]['y'] = y

    # and ensure asn is integer, x and y are floats

    for node in sorted(graph):
        graph.node[node]['asn'] = int(graph.node[node]['asn'])
        if graph.node[node]['asn'] == 0:
            log.debug('Node %s has ASN set to 0. Setting to 1'
                      % graph.node[node]['label'])
            graph.node[node]['asn'] = 1
        try:
            x = float(graph.node[node]['x'])
        except KeyError:
            x = 0
        graph.node[node]['x'] = x
        try:
            y = float(graph.node[node]['y'])
        except KeyError:
            y = 0
        graph.node[node]['y'] = y
        try:
            graph.node[node]['label']
        except KeyError:
            device_type = graph.node[node]['device_type']
            graph.node[node]['label'] = '%s_%s' \
                % (label_prefixes[device_type], unique_label.next())

    ank_edge_defaults = settings['Graphml']['Edge Defaults']
    edge_defaults = graph.graph['edge_default']
    for (key, val) in ank_edge_defaults.items():
        if key not in edge_defaults or edge_defaults[key] == 'None':
            edge_defaults[key] = val

    for (src, dst) in graph.edges():
        for (key, val) in edge_defaults.items():
            if key not in graph[src][dst]:
                graph[src][dst][key] = val

# apply defaults
# relabel nodes
# other handling... split this into seperate module!
# relabel based on label: assume unique by now!
    # if graph.graph.get("Network") == "European NRENs":
        # TODO: test if non-unique labels, if so then warn and proceed with this logic
        # we need to map node ids to contain network to ensure unique labels
        # mapping = dict( (n, "%s__%s" % (d['label'], d['asn'])) for n, d in graph.nodes(data=True))

    mapping = dict((n, d['label']) for (n, d) in graph.nodes(data=True))  # TODO: use dict comprehension
    if not all(key == val for (key, val) in mapping.items()):
        nx.relabel_nodes(graph, mapping, copy=False)  # Networkx wipes data if remap with same labels

    graph.graph['file_type'] = 'graphml'

    return graph



import os
from os.path import expanduser
import subprocess
import time
import shutil

cmd = ["python", "setup.py", "-V"]
child = subprocess.Popen(cmd, stdout=subprocess.PIPE)
version =  child.stdout.read().strip()

# Write the autonetkit version to a txt file to include in the autonetkit_cisco version
# (this is simpler than subclassing the setuptools module)

cmd = ["python", "setup.py", "clean", "bdist_wheel"]
"""
 Set WHEEL_TOOL env var as per
 https://wheel.readthedocs.org/en/latest/
 python setup.py bdist_wheel will automatically
 sign wheel files if the environment variable WHEEL_TOOL
 is set to the path of the wheel command line tool:
 """
# eg  $ export WHEEL_TOOL=/usr/local/bin/wheel
#TODO: look if other method to acheive this

new_env = os.environ.copy() # base on previous env, add new below:
new_env['WHEEL_TOOL'] = '/usr/local/bin/wheel'
child = subprocess.Popen(cmd, env = new_env, stdout=subprocess.PIPE)
for line in child.communicate():
    #print line
    pass

tar_file = "%s-%s-py27-none-any.whl" % ("autonetkit", version)
print tar_file



import autonetkit.log as log
from autonetkit.ank_utils import unwrap_graph, unwrap_nodes
import pprint
from collections import defaultdict

def igp_routes(anm, measured):
    #TODO: split up expected calculation and comparison so don't need to calculate SPF multiple times
    #TODO: allow choice of IGP - for now is just OSPF
    import networkx as nx
    g_ipv4 = anm['ipv4']
    g_ospf = anm['ospf']

    log.info("Verifying IGP routes")

    # extract source node from measured data
    try:
        src_node = measured[0][0]
    except IndexError:
        log.info("Unable to parse measured results, returning")
        return


    # calculate expected routes


    prefixes_by_router = {}
    prefix_reachability = defaultdict(list)
    for router in g_ospf:
        ipv4_node = g_ipv4.node(router)
        neighbors = ipv4_node.neighbors()
        neighbor_subnets = [n.subnet for n in neighbors]
        prefixes_by_router[router] = neighbor_subnets
        for subnet in neighbor_subnets:
            # use strings for faster matching from here on 
            # (can convert back to overlay_nodes later)
            prefix_reachability[str(subnet)].append(str(router))

    #print prefixes_by_router
    #print prefix_reachability

    graph = unwrap_graph(g_ospf)

    shortest_paths = nx.shortest_path(graph, source = src_node, weight = 'cost')
    shortest_path_lengths = nx.shortest_path_length(graph, source = src_node, weight = 'cost')
    #pprint.pprint(shortest_path_lengths)

    optimal_prefixes = {}
    for prefix, routers in prefix_reachability.items():
        # decorate with cost
        try:
            routers_with_costs = [(shortest_path_lengths[r], r) for r in routers]
        except KeyError:
            continue # no router, likely from eBGP
        min_cost = min(rwc[0] for rwc in routers_with_costs)
        shortest_routers = [rwc[1] for rwc in routers_with_costs if rwc[0] == min_cost]
        optimal_prefixes[str(prefix)] = shortest_routers
    #print optimal_prefixes

    verified_prefixes = {}
    for route in measured:
        dst_cd = route[-1]
        prefix = str(g_ipv4.node(dst_cd).subnet)
        try:
            optimal_routers = optimal_prefixes[prefix]
        except KeyError:
            continue # prefix not present
        if src_node in optimal_routers:
            continue # target is self

        optimal_routes = [shortest_paths[r] for r in optimal_routers]
        optimal_next_hop = [p[1] for p in optimal_routes]
        actual_next_hop = route[1]

        log.info( "Match: %s, %s, optimal: %s, actual: %s" % (
                actual_next_hop in optimal_next_hop, prefix, ", ".join(optimal_next_hop), actual_next_hop))

        if actual_next_hop not in optimal_next_hop:
            log.info("Unverified prefix: %s on %s. Expected: %s. Received: %s" % (prefix, dst_cd,
                ", ".join(optimal_next_hop), actual_next_hop))

        verified_prefixes[prefix] = actual_next_hop in optimal_next_hop


    verified_count = verified_prefixes.values().count(True) 
    try:
        verified_fraction = round(100 * verified_count/len(verified_prefixes),2)
    except ZeroDivisionError:
        verified_fraction = 0
    log.info("%s%% verification rate" % verified_fraction)
        
    return verified_prefixes



#!/usr/bin/python
# -*- coding: utf-8 -*-
import json

import autonetkit.ank as ank_utils
import autonetkit.ank_json
import autonetkit.ank_messaging
import autonetkit.log as log
import netaddr

# TODO: allow slack in allocations: both for ASN (group level), and for collision domains to allow new nodes to be easily added

try:
    import cPickle as pickle
except ImportError:
    import pickle


def assign_asn_to_interasn_cds(G_ip):
    #TODO: remove this if no longer needed

    # TODO: rename to assign_asn_to_cds as also does intra-asn cds
    # TODO: make this a common function to ip4 and ip6

    G_phy = G_ip.overlay('phy')
    for broadcast_domain in G_ip.nodes('broadcast_domain'):
        neigh_asn = list(ank_utils.neigh_attr(G_ip, broadcast_domain,
                         'asn', G_phy))  # asn of neighbors
        if len(set(neigh_asn)) == 1:
            asn = set(neigh_asn).pop()  # asn of any neigh, as all same
        else:
            asn = ank_utils.most_frequent(neigh_asn)  # allocate bc to asn with most neighbors in it
        broadcast_domain.asn = asn

    return

def allocate_loopbacks(g_ip, address_block=None):
    #TODO: handle no block specified
    loopback_blocks = {}
    loopback_pool = address_block.subnet(80)

    # consume the first address as it is the network address

    loopback_network = loopback_pool.next()  # network address

    unique_asns = set(n.asn for n in g_ip)
    for asn in sorted(unique_asns):
        loopback_blocks[asn] = loopback_pool.next()

    for (asn, devices) in g_ip.groupby('asn').items():
        loopback_hosts = loopback_blocks[asn].iter_hosts()
        loopback_hosts.next()  # drop .0 as a host address (valid but can be confusing)
        l3hosts = set(d for d in devices if d.is_l3device())
        for host in sorted(l3hosts, key=lambda x: x.label):
            host.loopback = loopback_hosts.next()

    g_ip.data.loopback_blocks = dict((asn, [subnet]) for (asn,
            subnet) in loopback_blocks.items())

def allocate_infra(g_ip, address_block=None):
    infra_blocks = {}

# TODO: check if need to do network address... possibly only for loopback_pool and infra_pool so maps to asn

    infra_pool = address_block.subnet(80)

    # consume the first address as it is the network address

    infra_network = infra_pool.next()  # network address

    unique_asns = set(n.asn for n in g_ip)
    for asn in sorted(unique_asns):
        infra_blocks[asn] = infra_pool.next()

    for (asn, devices) in sorted(g_ip.groupby('asn').items()):
        subnets = infra_blocks[asn].subnet(96)
        subnets.next()  # network address
        ptp_subnet = subnets.next().subnet(126)
        ptp_subnet.next()  # network address
        all_bcs = set(d for d in devices if d.broadcast_domain)
        ptp_bcs = [bc for bc in all_bcs if bc.degree() == 2]

        for bc in sorted(ptp_bcs):
            subnet = ptp_subnet.next()
            hosts = subnet.iter_hosts()
            hosts.next()  # drop .0 as a host address (valid but can be confusing)
            bc.subnet = subnet
            #TODO: check: should sort by default on dst as tie-breaker
            for edge in sorted(bc.edges(), key=lambda x: x.dst.label):
                edge.ip = hosts.next()

        non_ptp_cds = all_bcs - set(ptp_bcs)

        # break into /96 subnets

        for bc in sorted(non_ptp_cds):
            subnet = subnets.next()
            hosts = subnet.iter_hosts()
            hosts.next()  # drop .0 as a host address (valid but can be confusing)
            bc.subnet = subnet
            for edge in sorted(bc.edges(), key=lambda x: x.dst.label):
                edge.ip = hosts.next()


    g_ip.data.infra_blocks = dict((asn, [subnet]) for (asn, subnet) in
                                  infra_blocks.items())

def allocate_vrf_loopbacks(g_ip, address_block=None):
    secondary_loopback_blocks = {}
    secondary_loopback_pool = address_block.subnet(80)
    # consume the first address as it is the network address
    secondary_loopback_network = secondary_loopback_pool.next()  # network address

    unique_asns = set(n.asn for n in g_ip)
    for asn in sorted(unique_asns):
        secondary_loopback_blocks[asn] = \
            secondary_loopback_network.next()

    for (asn, devices) in g_ip.groupby('asn').items():
        l3hosts = set(d for d in devices if d.is_l3device())
        routers = [n for n in l3hosts if n.is_router()]  # filter
        secondary_loopbacks = [i for n in routers for i in
                               n.loopback_interfaces
                               if not i.is_loopback_zero]

        secondary_loopback_hosts = \
            secondary_loopback_blocks[asn].iter_hosts()
        secondary_loopback_hosts.next()  # drop .0 as a host address (valid but can be confusing)
        for interface in sorted(secondary_loopbacks):
            interface.loopback = secondary_loopback_hosts.next()
            interface.subnet = netaddr.IPNetwork("%s/128" % interface.loopback)

def allocate_ips(G_ip, infra_block = None, loopback_block = None, secondary_loopback_block = None):
    log.info('Allocating Host loopback IPs')
    #TODO: move the following step to the l3 graph
    assign_asn_to_interasn_cds(G_ip)

    allocate_loopbacks(G_ip, loopback_block)
    allocate_infra(G_ip, infra_block)
    allocate_vrf_loopbacks(G_ip, secondary_loopback_block)


import autonetkit.log as log
from autonetkit.collection.utils import get_results
import pkg_resources
import autonetkit.collection.process as ank_process
import autonetkit.ank_messaging as ank_messaging
import autonetkit
import json
import random
import autonetkit.console_script as console_script
import autonetkit.compilers.platform.netkit as pl_netkit
import os
import autonetkit.load.graphml as graphml
parse_template = pkg_resources.resource_filename("autonetkit", "textfsm/quagga/sh_ip_route")

emulation_server = "115.146.93.18"

anm = autonetkit.ANM()
anm.restore_latest()

nidb = autonetkit.NIDB()
nidb.restore_latest()
print list(nidb.nodes())

input_file = os.path.join("../../example/singleas.graphml")
input_graph = graphml.load_graphml(input_file)

import autonetkit.build_network as build_network
anm = build_network.initialise(input_graph)
anm = build_network.apply_design_rules(anm)

render_hostname = "localhost"

g_ospf = anm['ospf']
e1_8 = anm['ospf'].edge('5', '44')
print "before:", e1_8.cost
e1_8.cost = 1001
e1_8.apply_to_interfaces("cost")
print "after:", e1_8.cost

autonetkit.update_http(anm)

#nidb_a = nidb # copy original

nidb_b = console_script.create_nidb(anm)
render_hostname = "localhost"

platform_compiler = pl_netkit.NetkitCompiler(nidb_b, anm, render_hostname)
platform_compiler.compile()

autonetkit.update_http(anm, nidb_b)


import autonetkit.diff
diffs = autonetkit.diff.compare_nidb(nidb, nidb_b)
import pprint
pprint.pprint(diffs)


import autonetkit.push_changes
autonetkit.push_changes.apply_difference(nidb, diffs, emulation_server)




"""Console script entry point for AutoNetkit"""

import os
import random
import time
import sys
import traceback
from datetime import datetime

import autonetkit.ank_json as ank_json
import autonetkit.ank_messaging as ank_messaging
import autonetkit.config as config
import autonetkit.log as log
import autonetkit.render as render
import pkg_resources
from autonetkit import update_http
from autonetkit.nidb import NIDB

# TODO: make if measure set, then not compile - or warn if both set, as
# don't want to regen topology when measuring

try:
    ANK_VERSION = pkg_resources.get_distribution("autonetkit").version
except pkg_resources.DistributionNotFound:
    ANK_VERSION = "dev"


def file_monitor(filename):
    """Generator based function to check if a file has changed"""
    last_timestamp = os.stat(filename).st_mtime

    while True:
        timestamp = os.stat(filename).st_mtime
        if timestamp > last_timestamp:
            last_timestamp = timestamp
            yield True
        yield False


def manage_network(input_graph_string, timestamp, build_options, reload_build=False, grid=None):
    """Build, compile, render network as appropriate"""
    # import build_network_simple as build_network
    import autonetkit.build_network as build_network
    if reload_build:
# remap?
        build_network = reload(build_network)

    if build_options['build']:
        if input_graph_string:
            graph = build_network.load(input_graph_string)
        elif grid:
            graph = build_network.grid_2d(grid)

        anm = build_network.build(graph)
        if not build_options['compile']:
            update_http(anm)

        if build_options['validate']:
            import autonetkit.ank_validate
            try:
                autonetkit.ank_validate.validate(anm)
            except Exception, e:
                log.warning("Unable to validate topologies: %s" % e)

    if build_options['compile']:
        if build_options['archive']:
            anm.save()
        nidb = compile_network(anm)

        update_http(anm, nidb)
        log.debug("Sent ANM to web server")
        if build_options['archive']:
            nidb.save()
        # render.remove_dirs(["rendered"])
        if build_options['render']:
            render.render(nidb)

    if not(build_options['build'] or build_options['compile']):
        # Load from last run
        import autonetkit.anm
        anm = autonetkit.anm.AbstractNetworkModel()
        anm.restore_latest()
        nidb = NIDB()
        nidb.restore_latest()
        update_http(anm, nidb)

    if build_options['diff']:
        import autonetkit.diff
        nidb_diff = autonetkit.diff.nidb_diff()
        import json
        data = json.dumps(nidb_diff, cls=ank_json.AnkEncoder, indent=4)
        log.info("Wrote diff to diff.json")
        # TODO: make file specified in config
        with open("diff.json", "w") as fh:
            fh.write(data)

    if build_options['deploy']:
        deploy_network(anm, nidb, input_graph_string)

    if build_options['measure']:
        measure_network(anm, nidb)

    log.info("Finished")


def parse_options(argument_string=None):
    """Parse user-provided options"""
    import argparse
    usage = "autonetkit -f input.graphml"
    version = "%(prog)s " + str(ANK_VERSION)
    parser = argparse.ArgumentParser(description=usage, version=version)

    input_group = parser.add_mutually_exclusive_group()
    input_group.add_argument(
        '--file', '-f', default=None, help="Load topology from FILE")
    input_group.add_argument('--stdin', action="store_true", default=False,
                             help="Load topology from STDIN")

    parser.add_argument(
        '--monitor', '-m', action="store_true", default=False,
        help="Monitor input file for changes")
    parser.add_argument('--debug', action="store_true",
                        default=False, help="Debug mode")
    parser.add_argument('--quiet', action="store_true",
                        default=False, help="Quiet mode (only display warnings and errors)")
    parser.add_argument('--diff', action="store_true", default=False,
                        help="Diff NIDB")
    parser.add_argument('--compile', action="store_true",
                        default=False, help="Compile")
    parser.add_argument(
        '--build', action="store_true", default=False, help="Build")
    parser.add_argument(
        '--render', action="store_true", default=False, help="Compile")
    parser.add_argument(
        '--validate', action="store_true", default=False, help="Validate")
    parser.add_argument('--deploy', action="store_true",
                        default=False, help="Deploy")
    parser.add_argument('--archive', action="store_true", default=False,
                        help="Archive ANM, NIDB, and IP allocations")
    parser.add_argument('--measure', action="store_true",
                        default=False, help="Measure")
    parser.add_argument(
        '--webserver', action="store_true", default=False, help="Webserver")
    parser.add_argument('--grid', type=int, help="Grid Size (n * n)")
    parser.add_argument(
        '--target', choices=['netkit', 'cisco'], default=None)
    parser.add_argument(
        '--vis_uuid', default=None, help="UUID for multi-user visualisation")
    if argument_string:
        arguments = parser.parse_args(argument_string.split())
    else:
        # from command line arguments
        arguments = parser.parse_args()
    return arguments


def main(options):
    settings = config.settings

    if options.vis_uuid:
        config.settings['Http Post']['uuid'] = options.vis_uuid

    try:
        # test if can import, if not present will fail and not add to template
        # path
        import autonetkit_cisco
    except ImportError:
        pass
    else:
        import autonetkit_cisco.version
        version_banner = autonetkit_cisco.version.banner()
        log.info("%s" % version_banner)

    log.info("AutoNetkit %s" % ANK_VERSION)

    if options.target == "cisco":
        # output target is Cisco
        log.info("Setting output target as Cisco")
        settings['Graphml']['Node Defaults']['platform'] = "VIRL"
        settings['Graphml']['Node Defaults']['host'] = "internal"
        settings['Graphml']['Node Defaults']['syntax'] = "ios_xr"
        settings['Compiler']['Cisco']['to memory'] = 1
        settings['General']['deploy'] = 1
        settings['Deploy Hosts']['internal'] = {'VIRL':
                                                {'deploy': 1}}

    if options.debug or settings['General']['debug']:
        # TODO: fix this
        import logging
        logger = logging.getLogger("ANK")
        logger.setLevel(logging.DEBUG)

    if options.quiet or settings['General']['quiet']:
        import logging
        logger = logging.getLogger("ANK")
        logger.setLevel(logging.WARNING)

    build_options = {
        'compile': options.compile or settings['General']['compile'],
        'render': options.render or settings['General']['render'],
        'validate': options.validate or settings['General']['validate'],
        'build': options.build or settings['General']['build'],
        'deploy': options.deploy or settings['General']['deploy'],
        'measure': options.measure or settings['General']['measure'],
        'monitor': options.monitor or settings['General']['monitor'],
        'diff': options.diff or settings['General']['diff'],
        'archive': options.archive or settings['General']['archive'],
    }

    if options.webserver:
        log.info("Webserver not yet supported, please run as seperate module")

    if options.file:
        with open(options.file, "r") as fh:
            input_string = fh.read()
        timestamp = os.stat(options.file).st_mtime
    elif options.stdin:
        import sys
        input_string = sys.stdin
        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S_%f")
    elif options.grid:
        input_string = ""
        now = datetime.now()
        timestamp = now.strftime("%Y%m%d_%H%M%S_%f")
    else:
        log.info("No input file specified. Exiting")
        raise SystemExit

    try:
        manage_network(input_string, timestamp,
                       build_options=build_options, grid=options.grid)
    except Exception, err:
        log.error(
            "Error generating network configurations: %s. More information may be available in the debug log." % err)
        log.debug("Error generating network configurations", exc_info=True)
        if settings['General']['stack_trace']:
            print traceback.print_exc()
        import sys
        sys.exit("Unable to build configurations.")

# TODO: work out why build_options is being clobbered for monitor mode
    build_options['monitor'] = options.monitor or settings['General'][
        'monitor']

    if build_options['monitor']:
        try:
            log.info("Monitoring for updates...")
            input_filemonitor = file_monitor(options.file)
            #build_filemonitor = file_monitor("autonetkit/build_network.py")
            while True:
                time.sleep(1)
                rebuild = False
                reload_build = False
                if input_filemonitor.next():
                    rebuild = True
                # if build_filemonitor.next():
                    #reload_build = True
                    #rebuild = True

                if rebuild:
                    try:
                        log.info("Input graph updated, recompiling network")
                        with open(options.file, "r") as fh:
                            input_string = fh.read()  # read updates
                        manage_network(input_string,
                                       timestamp, build_options, reload_build)
                        log.info("Monitoring for updates...")
                    except Exception, e:
                        log.warning("Unable to build network %s" % e)
                        traceback.print_exc()

        except KeyboardInterrupt:
            log.info("Exiting")


def create_nidb(anm):
    nidb = NIDB()
    g_phy = anm['phy']
    g_ip = anm['ip']
    g_graphics = anm['graphics']
    nidb.add_nodes_from(
        g_phy, retain=['label', 'host', 'platform', 'Network', 'update', 'asn'])

    cd_nodes = [n for n in g_ip.nodes(
        "broadcast_domain") if not n.is_switch()]  # Only add created cds - otherwise overwrite host of switched
    nidb.add_nodes_from(
        cd_nodes, retain=['label', 'host'], broadcast_domain=True)

    for node in nidb.nodes("broadcast_domain"):
        ipv4_node = anm['ipv4'].node(node)
        if ipv4_node:
            node.ipv4_subnet = ipv4_node.subnet
            #TODO: copy across IPv6 seperately
            node.ipv6_subnet = ipv4_node['ipv6'].subnet

# add edges to switches
    edges_to_add = [edge for edge in g_phy.edges()
                    if edge.src.is_switch() or edge.dst.is_switch()]
    # cd edges from split
    edges_to_add += [edge for edge in g_ip.edges() if edge.split]
    nidb.add_edges_from(edges_to_add)

    nidb.copy_graphics(g_graphics)

    return nidb


def compile_network(anm):
    nidb = create_nidb(anm)
    g_phy = anm['phy']

    for target, target_data in config.settings['Compile Targets'].items():
        host = target_data['host']
        platform = target_data['platform']
        if platform == "netkit":
            import autonetkit.compilers.platform.netkit as pl_netkit
            platform_compiler = pl_netkit.NetkitCompiler(nidb, anm, host)
        elif platform == "VIRL":
            try:
                import autonetkit_cisco.compilers.platform.cisco as pl_cisco
                platform_compiler = pl_cisco.CiscoCompiler(nidb, anm, host)
            except ImportError, e:
                log.debug("Unable to load VIRL platform compiler")
        elif platform == "dynagen":
            import autonetkit.compilers.platform.dynagen as pl_dynagen
            platform_compiler = pl_dynagen.DynagenCompiler(nidb, anm, host)
        elif platform == "junosphere":
            import autonetkit.compilers.platform.junosphere as pl_junosphere
            platform_compiler = pl_junosphere.JunosphereCompiler(
                nidb, anm, host)

        if any(g_phy.nodes(host=host, platform=platform)):
            log.info("Compiling configurations for %s on %s" % (platform, host))
            platform_compiler.compile()  # only compile if hosts set
        else:
            log.debug("No devices set for %s on %s" % (platform, host))

    return nidb


def deploy_network(anm, nidb, input_graph_string=None):

    log.info("Deploying Network")

    deploy_hosts = config.settings['Deploy Hosts']
    for hostname, host_data in deploy_hosts.items():
        for platform, platform_data in host_data.items():
            if not any(nidb.nodes(host=hostname, platform=platform)):
                log.debug("No hosts for (host, platform) (%s, %s), skipping deployment"
                          % (hostname, platform))
                continue

            if not platform_data['deploy']:
                log.debug("Not deploying to %s on %s" % (platform, hostname))
                continue

            config_path = os.path.join("rendered", hostname, platform)

            if hostname == "internal":
                try:
                    from autonetkit_cisco import deploy as cisco_deploy
                except ImportError:
                    pass  # development module, may not be available
                if platform == "VIRL":
                    create_new_xml = False
                    if not input_graph_string:
                        create_new_xml = True  # no input, eg if came from grid
                    elif anm['input'].data['file_type'] == "graphml":
                        create_new_xml = True  # input from graphml, create XML

                    if create_new_xml:
                        cisco_deploy.create_xml(anm, nidb, input_graph_string)
                    else:
                        cisco_deploy.package(nidb, config_path,
                                             input_graph_string)
                continue

            username = platform_data['username']
            key_file = platform_data['key_file']
            host = platform_data['host']

            if platform == "netkit":
                import autonetkit.deploy.netkit as netkit_deploy
                tar_file = netkit_deploy.package(config_path, "nklab")
                netkit_deploy.transfer(
                    host, username, tar_file, tar_file, key_file)
                netkit_deploy.extract(host, username, tar_file,
                                      config_path, timeout=60, key_filename=key_file,
                                      parallel_count=10)
                if platform == "VIRL":
                    # TODO: check why using nklab here
                    cisco_deploy.package(config_path, "nklab")


def console_entry():
    """If come from console entry point"""
    args = parse_options()
    main(args)

if __name__ == "__main__":
    try:
        args = parse_options()
        main(args)
    except KeyboardInterrupt:
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
import autonetkit.log as log
import autonetkit.ank as ank_utils

import itertools
from autonetkit.ank_utils import call_log

@call_log
def mpls_te(anm):
    g_in = anm['input']
    g_phy = anm['phy']
    g_l3_conn = anm['l3_conn']

    # add regardless, so allows quick check of node in anm['mpls_te'] in compilers

    g_mpls_te = anm.add_overlay('mpls_te')
    if not any(True for n in g_in.routers() if n.mpls_te_enabled):
        log.debug('No nodes with mpls_te_enabled set')
        return

    # te head end set if here

    g_mpls_te.add_nodes_from(g_in.routers())

    # build up edge list sequentially, to provide meaningful messages for multipoint links

    multipoint_edges = [e for e in g_l3_conn.edges() if e.multipoint]
    if len(multipoint_edges):
        log.info('Excluding multi-point edges from MPLS TE topology: %s'
           % ', '.join(str(e) for e in multipoint_edges))

    edges_to_add = set(g_l3_conn.edges()) - set(multipoint_edges)
    g_mpls_te.add_edges_from(edges_to_add)


@call_log
def mpls_oam(anm):
    g_in = anm['input']

    # create placeholder graph (may have been created in other steps)

    if anm.has_overlay('mpls_oam'):
        g_mpls_oam = anm['mpls_oam']
    else:
        g_mpls_oam = anm.add_overlay('mpls_oam')

    use_mpls_oam = g_in.data.use_mpls_oam
    if use_mpls_oam:
        g_mpls_oam.add_nodes_from(g_in.routers())

@call_log
def vrf_pre_process(anm):
    """Marks nodes in g_in as appropriate based on vrf roles.
    CE nodes -> ibgp_role = Disabled, so not in iBGP (this is allocated later)
    """
    log.debug("Applying VRF pre-processing")
    g_vrf = anm['vrf']
    for node in g_vrf.nodes(vrf_role = "CE"):
        log.debug("Marking CE node %s as non-ibgp" % node)
        node['input'].ibgp_role = "Disabled"

@call_log
def allocate_vrf_roles(g_vrf):
    """Allocate VRF roles"""
    g_phy = g_vrf.anm['phy']
    #TODO: might be clearer like ibgp with is_p is_pe etc booleans?  - final step to translate to role for vis
    for node in g_vrf.nodes(vrf_role="CE"):
        if not node.vrf:
            node.vrf = "default_vrf"

    ce_set_nodes = []
    for node in sorted(g_vrf.nodes('vrf')):
        node.vrf_role = "CE"
        ce_set_nodes.append(node)
    if len(ce_set_nodes):
        message = ", ".join(str(n) for n in sorted(ce_set_nodes))
        g_vrf.log.info("VRF role set to CE for %s" % message)

    non_ce_nodes = [node for node in g_vrf if node.vrf_role != "CE"]

    pe_set_nodes = []
    p_set_nodes = []
    for node in sorted(non_ce_nodes):
        phy_neighbors = [n for n in g_phy.node(node).neighbors() if n.is_router()]
        # neighbors from physical graph for connectivity
        #TODO: does this do anything?
        phy_neighbors = [neigh for neigh in phy_neighbors]
            # filter to just this asn
        if any(g_vrf.node(neigh).vrf_role == "CE" for neigh in phy_neighbors):
            # phy neigh has vrf set in this graph
            node.vrf_role = "PE"
            pe_set_nodes.append(node)
        else:
            node.vrf_role = "P"  # default role
            p_set_nodes.append(node)

    if len(pe_set_nodes):
        message = ", ".join(str(n) for n in sorted(pe_set_nodes))
        g_vrf.log.info("VRF role set to PE for %s" % message)
    if len(p_set_nodes):
        message = ", ".join(str(n) for n in sorted(p_set_nodes))
        g_vrf.log.info("VRF role set to P for %s" % message)

@call_log
def add_vrf_loopbacks(g_vrf):
    """Adds loopbacks for VRFs, and stores VRFs connected to PE router"""
    #autonetkit.update_http(anm)
    for node in g_vrf.nodes(vrf_role="PE"):
        node_vrf_names = {n.vrf for n in node.neighbors(vrf_role="CE")}
        node.node_vrf_names = node_vrf_names
        node.rd_indices = {}
        for index, vrf_name in enumerate(node_vrf_names, 1):
            node.rd_indices[vrf_name] = index
            node.add_loopback(vrf_name=vrf_name,
                              description="loopback for vrf %s" % vrf_name)

@call_log
def build_ibgp_vpn_v4(anm):
    """Based on the ibgp_v4 hierarchy rules.
    Exceptions:
    1. Remove links to (PE, RRC) nodes

    CE nodes are excluded from RR hierarchy ibgp creation through pre-process step

    """
    #TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    g_bgp = anm['bgp']
    g_ibgp_v4 = anm['ibgp_v4']
    g_vrf = anm['vrf']
    g_ibgp_vpn_v4 = anm.add_overlay("ibgp_vpn_v4", directed=True)

    v6_vrf_nodes = [n for n in g_vrf
        if n.vrf is not None and n['phy'].use_ipv6 is True]
    if len(v6_vrf_nodes):
        message = ", ".join(str(s) for s in v6_vrf_nodes)
        log.warning("This version of AutoNetkit does not support IPv6 MPLS VPNs. "
            "The following nodes have IPv6 enabled but will not have an associated IPv6 MPLS VPN topolog created: %s" % message)

    ibgp_v4_nodes = list(g_ibgp_v4.nodes())
    pe_nodes = set(g_vrf.nodes(vrf_role = "PE"))
    pe_rrc_nodes = {n for n in ibgp_v4_nodes if
            n in pe_nodes and n.ibgp_role == "RRC"}
    #TODO: warn if pe_rrc_nodes?
    ce_nodes = set(g_vrf.nodes(vrf_role = "CE"))

    if len(pe_nodes) == len(ce_nodes) == len(pe_rrc_nodes) == 0:
        # no vrf nodes to connect
        return

    #TODO: extend this to only connect nodes which are connected in VRFs, so don't set to others

    ibgp_vpn_v4_nodes = (n for n in ibgp_v4_nodes
            if n not in ce_nodes)
    g_ibgp_vpn_v4.add_nodes_from(ibgp_vpn_v4_nodes, retain = ["ibgp_role"])
    g_ibgp_vpn_v4.add_edges_from(g_ibgp_v4.edges(), retain = "direction")

    for node in g_ibgp_vpn_v4:
        if node.ibgp_role in ("HRR", "RR"):
            node.retain_route_target = True

    ce_edges = [e for e in g_ibgp_vpn_v4.edges()
            if e.src in ce_nodes or e.dst in ce_nodes]

    # mark ibgp direction
    ce_pe_edges = []
    pe_ce_edges = []
    for edge in g_ibgp_vpn_v4.edges():
        if (edge.src.vrf_role, edge.dst.vrf_role) == ("CE", "PE"):
            edge.direction = "up"
            edge.vrf = edge.src.vrf
            ce_pe_edges.append(edge)
        elif (edge.src.vrf_role, edge.dst.vrf_role) == ("PE", "CE"):
            edge.direction = "down"
            edge.vrf = edge.dst.vrf
            pe_ce_edges.append(edge)

    #TODO: Document this
    g_ibgpv4 = anm['ibgp_v4']
    g_ibgpv6 = anm['ibgp_v6']
    g_ibgpv4.remove_edges_from(ce_edges)
    g_ibgpv6.remove_edges_from(ce_edges)
    g_ibgpv4.add_edges_from(ce_pe_edges, retain = ["direction", "vrf"])
    g_ibgpv4.add_edges_from(pe_ce_edges, retain = ["direction", "vrf"])
    g_ibgpv6.add_edges_from(ce_pe_edges, retain = ["direction", "vrf"])
    g_ibgpv6.add_edges_from(pe_ce_edges, retain = ["direction", "vrf"])
    for edge in pe_ce_edges:
        # mark as exclude so don't include in standard ibgp config stanzas
        if g_ibgpv4.has_edge(edge):
            edge['ibgp_v4'].exclude = True
        if g_ibgpv6.has_edge(edge):
            edge['ibgp_v6'].exclude = True

# legacy
    g_bgp = anm['bgp']
    g_bgp.remove_edges_from(ce_edges)
    g_bgp.add_edges_from(ce_pe_edges, retain = ["direction", "vrf", "type"])
    g_bgp.add_edges_from(pe_ce_edges, retain = ["direction", "vrf", "type"])

    # also need to modify the ibgp_v4 and ibgp_v6 graphs

@call_log
def build_mpls_ldp(anm):
    """Builds MPLS LDP"""
    g_in = anm['input']
    g_vrf = anm['vrf']
    g_l3conn = anm['l3_conn']
    g_mpls_ldp = anm.add_overlay("mpls_ldp")
    nodes_to_add = [n for n in g_in.routers()
            if n['vrf'].vrf_role in ("PE", "P")]
    g_mpls_ldp.add_nodes_from(nodes_to_add, retain=["vrf_role", "vrf"])

    # store as set for faster lookup
    pe_nodes = set(g_vrf.nodes(vrf_role = "PE"))
    p_nodes = set(g_vrf.nodes(vrf_role = "P"))

    pe_to_pe_edges = (e for e in g_l3conn.edges()
            if e.src in pe_nodes and e.dst in pe_nodes)
    g_mpls_ldp.add_edges_from(pe_to_pe_edges)

    pe_to_p_edges = (e for e in g_l3conn.edges()
            if e.src in pe_nodes and e.dst in p_nodes
            or e.src in p_nodes and e.dst in pe_nodes)
    g_mpls_ldp.add_edges_from(pe_to_p_edges)

    p_to_p_edges = (e for e in g_l3conn.edges()
            if e.src in p_nodes and e.dst in p_nodes)
    g_mpls_ldp.add_edges_from(p_to_p_edges)

@call_log
def mark_ebgp_vrf(anm):
    g_ebgp = anm['ebgp']
    g_vrf = anm['vrf']
    g_ebgpv4 = anm['ebgp_v4']
    g_ebgpv6 = anm['ebgp_v6']
    pe_nodes = set(g_vrf.nodes(vrf_role = "PE"))
    ce_nodes = set(g_vrf.nodes(vrf_role = "CE"))
    for edge in g_ebgpv4.edges():
        if (edge.src in pe_nodes and edge.dst in ce_nodes):
            # exclude from "regular" ebgp (as put into vrf stanza)
            edge.exclude = True
            edge.vrf = edge.dst['vrf'].vrf

    for edge in g_ebgpv6.edges():
        if (edge.src in pe_nodes and edge.dst in ce_nodes):
             # exclude from "regular" ebgp (as put into vrf stanza)
            edge.exclude = True
            edge.vrf = edge.dst['vrf'].vrf

@call_log
def build_vrf(anm):
    """Build VRF Overlay"""
    g_in = anm['input']
    g_l3conn = anm['l3_conn']
    g_vrf = anm.add_overlay("vrf")

    import autonetkit
    autonetkit.ank.set_node_default(g_in,  vrf=None)

    if not any(True for n in g_in.routers() if n.vrf):
        log.debug("No VRFs set")
        return

    g_vrf.add_nodes_from(g_in.routers(), retain=["vrf_role", "vrf"])

    allocate_vrf_roles(g_vrf)
    vrf_pre_process(anm)

    def is_pe_ce_edge(edge):
        src_vrf_role = g_vrf.node(edge.src).vrf_role
        dst_vrf_role = g_vrf.node(edge.dst).vrf_role
        return (src_vrf_role, dst_vrf_role) in (("PE", "CE"), ("CE", "PE"))

    vrf_add_edges = (e for e in g_l3conn.edges()
           if is_pe_ce_edge(e))
    #TODO: should mark as being towards PE or CE
    g_vrf.add_edges_from(vrf_add_edges)

    def is_pe_p_edge(edge):
        src_vrf_role = g_vrf.node(edge.src).vrf_role
        dst_vrf_role = g_vrf.node(edge.dst).vrf_role
        return (src_vrf_role, dst_vrf_role) in (("PE", "P"), ("P", "PE"))
    vrf_add_edges = (e for e in g_l3conn.edges()
            if is_pe_p_edge(e))
    g_vrf.add_edges_from(vrf_add_edges)

    build_mpls_ldp(anm)
    # add PE to P edges

    add_vrf_loopbacks(g_vrf)
    # allocate route-targets per AS
    # This could later look at connected components for each ASN
    route_targets = {}
    for asn, devices in ank_utils.groupby("asn", g_vrf.nodes(vrf_role = "PE")):
        asn_vrfs = [d.node_vrf_names for d in devices]
        # flatten list to unique set
        asn_vrfs = set(itertools.chain.from_iterable(asn_vrfs))
        route_targets[asn] = {vrf: "%s:%s" % (asn, index)
                for index, vrf in enumerate(sorted(asn_vrfs), 1)}

    g_vrf.data.route_targets = route_targets

    for node in g_vrf:
        vrf_loopbacks = node.interfaces("is_loopback", "vrf_name")
        for index, interface in enumerate(vrf_loopbacks, start = 101):
            interface.index = index

    for edge in g_vrf.edges():
        # Set the vrf of the edge to be that of the CE device (either src or dst)
        edge.vrf = edge.src.vrf if edge.src.vrf_role is "CE" else edge.dst.vrf

    # map attributes to interfaces
    for edge in g_vrf.edges():
        for interface in edge.interfaces():
            interface.vrf_name = edge.vrf


#!/usr/bin/python
# -*- coding: utf-8 -*-
import functools
import itertools
import json
import math
import time
from collections import defaultdict

import autonetkit.ank as ank_utils
import autonetkit.ank_json
import autonetkit.ank_messaging
import autonetkit.log as log
import netaddr
import networkx as nx
from autonetkit.exception import AutoNetkitException

# TODO: allow slack in allocations: both for ASN (group level), and for collision domains to allow new nodes to be easily added

try:
    import cPickle as pickle
except ImportError:
    import pickle


def subnet_size(host_count):
    """Returns subnet size"""

    host_count += 2  # network and broadcast
    return int(math.ceil(math.log(host_count, 2)))


class TreeNode(object):

    def __init__(self, graph, node):
        object.__setattr__(self, 'graph', graph)
        object.__setattr__(self, 'node', node)

# TODO: make thise fixed attributes, as only certain number needed here

    def __getattr__(self, attr):
        return self.graph.node[self.node].get(attr)

    def __setattr__(self, key, val):
        self.graph.node[self.node][key] = val

    def __lt__(self, other):
        if self.host and other.host:
            return self.host < other.host
        return self.node < other.node

# TODO: restore function that truncated subnets

    def __repr__(self):
        if self.host:
            return '%s %s' % (self.subnet, self.host)
        if self.loopback_group:
            return 'Lo Gr %s: %s' % (self.group_attr, self.subnet)
        if self.group_attr:
            return '%s: %s' % (self.group_attr, self.subnet)
        if self.subnet:
            return '%s' % self.subnet
        return 'TreeNode: %s' % self.node

    def is_broadcast_domain(self):
        return self.host and self.host.broadcast_domain

    def is_loopback_group(self):
        return self.loopback_group

    def is_interface(self):
        return isinstance(self.host, autonetkit.anm.overlay_interface)

    def is_host(self):
        return bool(self.host)

    def children(self):
        return [TreeNode(self.graph, child) for child in
                self.graph.successors(self.node)]


class IpTree(object):

    def __init__(self, root_ip_block):
        self.unallocated_nodes = []
        self.graph = nx.DiGraph()
        self.root_node = None
        self.timestamp = time.strftime('%Y%m%d_%H%M%S',
                time.localtime())

# taken_nodes -> these are nodes manually specified, eg in graphml

        self.node_id_counter = (i for i in itertools.count(0) if i
            not in self.graph)
        self.root_ip_block = root_ip_block

    def __len__(self):
        return len(self.graph)

    def __iter__(self):
        return iter(TreeNode(self.graph, node) for node in self.graph)

    @property
    def next_node_id(self):
        return self.node_id_counter.next()

    def add_parent_nodes(self, subgraph, level_counts):
        for level in range(32, 0, -1):
            try:
                current_count = float(level_counts[level])  # float so do floating point division
            except KeyError:
                continue  # key not present - likely higher up tree
            parent_count = int(math.ceil(current_count / 2))
            parent_level = level - 1
            level_counts[parent_level] += parent_count
            subgraph.add_nodes_from((self.next_node_id,
                                    {'prefixlen': parent_level})
                                    for n in range(parent_count))

            if level_counts[parent_level] == 1:
                if parent_level == min(level_counts.keys()):
                    children = [n for n in subgraph
                                if subgraph.node[n]['prefixlen']
                                == parent_level + 1]
                    if all('host' in subgraph.node[n] for n in
                           children):
                        subgraph.add_node(self.next_node_id,
                                prefixlen=parent_level - 1)

                    break  # Reached top of tree

    def build_tree(
        self,
        subgraph,
        level_counts,
        nodes_by_level,
        ):
        smallest_prefix = min(level_counts.keys())
        for prefixlen in range(smallest_prefix, 32):

            # TODO: fix sorting here

            unallocated_children = set(nodes_by_level[prefixlen + 1])
            for node in sorted(nodes_by_level[prefixlen]):
                is_not_subnet = not ('host' in subgraph.node[node]
                        or 'group_attr' in subgraph.node[node])
                if is_not_subnet:
                    child_a = unallocated_children.pop()
                    subgraph.add_edge(node, child_a)
                    try:
                        child_b = unallocated_children.pop()
                        subgraph.add_edge(node, child_b)
                    except KeyError:
                        pass  # single child, just attach

        root_node = nodes_by_level[smallest_prefix][0]
        return root_node

    def save(self):
        import os
        archive_dir = os.path.join('versions', 'ip')
        if not os.path.isdir(archive_dir):
            os.makedirs(archive_dir)

        self.graph.graph['timestamp'] = self.timestamp
        data = autonetkit.ank_json.ank_json_dumps(self.graph)

# TODO: should this use the ank_json.jsonify_nidb() ?

        json_file = 'ip_%s.json.gz' % self.timestamp
        json_path = os.path.join(archive_dir, json_file)
        log.debug('Saving to %s' % json_path)

        # with gzip.open(json_path, "wb") as json_fh:

        with open(json_path, 'wb') as json_fh:
            json_fh.write(data)

    def build(self, group_attr='asn'):
        """Builds tree from unallocated_nodes,
        groupby is the attribute to build subtrees from"""

        subgraphs = []

# if network final octet is .0 eg 10.0.0.0 or 192.168.0.0, then add extra "dummy" node, so don't have a loopback of 10.0.0.0
# Change strategy: if just hosts (ie loopbacks), then allocate as a large collision domain

        if not len(self.unallocated_nodes):

# no nodes to allocate - eg could be no collision domains

            return

        unallocated_nodes = self.unallocated_nodes
        key_func = lambda x: x.get(group_attr)
        if all(isinstance(item, autonetkit.anm.overlay_interface)
               and item.is_loopback for item in unallocated_nodes):
            key_func = lambda x: x.node.get(group_attr)  # interface, map key function to be the interface's node

        unallocated_nodes = sorted(unallocated_nodes, key=key_func)
        groupings = itertools.groupby(unallocated_nodes, key=key_func)
        prefixes_by_attr = {}

        for (attr_value, items) in groupings:

# make subtree for each attr

            items = sorted(list(items))
            subgraph = nx.DiGraph()

            if all(isinstance(item, autonetkit.anm.overlay_interface)
                   for item in items):

                # interface

                if all(item.is_loopback for item in items):
                    parent_id = self.next_node_id
                    prefixlen = 32 - subnet_size(len(items))  # group all loopbacks into single subnet
                    subgraph.add_node(parent_id, prefixlen=prefixlen,
                            loopback_group=True)
                    for item in items:

                    # subgraph.add_edge(node, child_a)

                        item_id = self.next_node_id
                        subgraph.add_node(item_id, prefixlen=32,
                                host=item)
                        subgraph.add_edge(parent_id, item_id)

                    root_node = parent_id
                    subgraphs.append(subgraph)
                    subgraph.graph['root'] = root_node
                    subgraph.node[root_node]['group_attr'] = attr_value
                    continue  # finished for loopbacks, continue only for collision domains

            if all(item.is_l3device() for item in items):

                # Note: only l3 devices are added for loopbacks: cds allocate to edges not devices (for now) - will be fixed when move to proper interface model

                parent_id = self.next_node_id
                prefixlen = 32 - subnet_size(len(items))  # group all loopbacks into single subnet
                subgraph.add_node(parent_id, prefixlen=prefixlen,
                                  loopback_group=True)
                for item in items:

                # subgraph.add_edge(node, child_a)

                    item_id = self.next_node_id
                    subgraph.add_node(item_id, prefixlen=32, host=item)
                    subgraph.add_edge(parent_id, item_id)

                root_node = parent_id
                subgraphs.append(subgraph)
                subgraph.graph['root'] = root_node
                subgraph.node[root_node]['group_attr'] = attr_value
                continue  # finished for loopbacks, continue only for collision domains

            for item in items:
                if item.broadcast_domain:
                    subgraph.add_node(self.next_node_id, prefixlen=32
                            - subnet_size(item.degree()), host=item)
                if item.is_l3device():
                    subgraph.add_node(self.next_node_id, prefixlen=32,
                            host=item)

            # now group by levels

            level_counts = defaultdict(int)

            nodes_by_level = defaultdict(list)
            for node in subgraph.nodes():
                prefixlen = subgraph.node[node]['prefixlen']
                nodes_by_level[prefixlen].append(node)

            log.debug('Building IP subtree for %s %s' % (group_attr,
                      attr_value))

            for (level, nodes) in nodes_by_level.items():
                level_counts[level] = len(nodes)

            self.add_parent_nodes(subgraph, level_counts)

# test if min_level node is bound, if so then add a parent, so root for AS isn't a cd

            min_level = min(level_counts)
            min_level_nodes = [n for n in subgraph
                               if subgraph.node[n]['prefixlen']
                               == min_level]

            # test if bound

            if len(min_level_nodes) == 2:
                subgraph.add_node(self.next_node_id,
                                  {'prefixlen': min_level - 2})
                subgraph.add_node(self.next_node_id,
                                  {'prefixlen': min_level - 2})
                subgraph.add_node(self.next_node_id,
                                  {'prefixlen': min_level - 1})
            if len(min_level_nodes) == 1:
                subgraph.add_node(self.next_node_id,
                                  {'prefixlen': min_level - 1})

            # rebuild with parent nodes

            nodes_by_level = defaultdict(list)
            for node in subgraph.nodes():
                prefixlen = subgraph.node[node]['prefixlen']
                nodes_by_level[prefixlen].append(node)

            root_node = self.build_tree(subgraph, level_counts,
                    nodes_by_level)
            subgraphs.append(subgraph)

            subgraph.graph['root'] = root_node

# FOrce to be a /16 block
# TODO: document this

            subgraph.node[root_node]['prefixlen'] = 16
            subgraph.node[root_node]['group_attr'] = attr_value
            prefixes_by_attr[attr_value] = subgraph.node[root_node]['prefixlen']

        global_graph = nx.DiGraph()
        subgraphs = sorted(subgraphs, key=lambda x: \
                           subgraph.node[subgraph.graph['root'
                           ]]['group_attr'])
        root_nodes = [subgraph.graph['root'] for subgraph in subgraphs]
        root_nodes = []
        for subgraph in subgraphs:
            root_node = subgraph.graph['root']
            root_nodes.append(root_node)
            global_graph.add_node(root_node, subgraph.node[root_node])

        nodes_by_level = defaultdict(list)
        for node in root_nodes:
            prefixlen = global_graph.node[node]['prefixlen']
            nodes_by_level[prefixlen].append(node)

        level_counts = defaultdict(int)
        for (level, nodes) in nodes_by_level.items():
            level_counts[level] = len(nodes)

        self.add_parent_nodes(global_graph, level_counts)

# rebuild nodes by level
# TODO: make this a function

        nodes_by_level = defaultdict(list)
        for node in global_graph:
            prefixlen = global_graph.node[node]['prefixlen']
            nodes_by_level[prefixlen].append(node)

        global_root = self.build_tree(global_graph, level_counts,
                nodes_by_level)
        global_root = TreeNode(global_graph, global_root)

        for subgraph in subgraphs:
            global_graph = nx.compose(global_graph, subgraph)

        # now allocate the IPs

        global_prefix_len = global_root.prefixlen

        # TODO: try/catch if the block is too small for prefix

        try:
            global_ip_block = \
                self.root_ip_block.subnet(global_prefix_len).next()
        except StopIteration:
            #message = ("Unable to allocate IPv4 subnets. ")
            formatted_prefixes = ", ".join("AS%s: /%s" % (k,v) for k,v in sorted(prefixes_by_attr.items()))
            message = ("Cannot create requested number of /%s subnets from root block %s. Please specify a larger root IP block. (Requested subnet allocations are: %s)"
                       % (global_prefix_len, self.root_ip_block, formatted_prefixes))
            log.error(message)
            raise AutoNetkitException(message)  # TODO: throw ANK specific exception here
        self.graph = global_graph

# add children of collision domains

        cd_nodes = [n for n in self if n.is_broadcast_domain()]
        for cd in cd_nodes:
            for edge in sorted(cd.host.edges()):

                # TODO: sort these

                child_id = self.next_node_id
                cd_id = cd.node
                global_graph.add_node(child_id, prefixlen=32,
                        host=edge.dst_int)
                global_graph.add_edge(cd_id, child_id)  # cd -> neigh (cd is parent)

# TODO: make allocate seperate step

        def allocate(node):

            # children = graph.successors(node)

            children = sorted(node.children())
            prefixlen = node.prefixlen + 1

            # workaround for clobbering attr subgraph root node with /16 if was a /28

            subnet = node.subnet.subnet(prefixlen)

# handle case where children subnet

            if node.is_loopback_group() or node.is_broadcast_domain():  # special case of single AS -> root is loopback_group

                # TODO: generalise this rather than repeated code with below
                # node.subnet = subnet.next() # Note: don't break into smaller subnets if single-AS

                iterhosts = node.subnet.iter_hosts()  # ensures start at .1 rather than .0
                sub_children = node.children()
                for sub_child in sub_children:

                    # TODO: tidy up this allocation to always record the subnet

                    if sub_child.is_interface() \
                        and sub_child.host.is_loopback:
                        if sub_child.host.is_loopback_zero:

                            # loopback zero, just store the ip address

                            sub_child.ip_address = iterhosts.next()
                        else:

                            # secondary loopback

                            sub_child.ip_address = iterhosts.next()
                            sub_child.subnet = node.subnet
                    elif sub_child.is_interface() \
                        and sub_child.host.is_physical:

                        # physical interface

                        sub_child.ip_address = iterhosts.next()
                        sub_child.subnet = node.subnet
                    else:
                        sub_child.subnet = iterhosts.next()

                return

            for child in children:

                # traverse the tree

                if child.is_broadcast_domain():
                    subnet = subnet.next()
                    child.subnet = subnet
                    iterhosts = child.subnet.iter_hosts()  # ensures start at .1 rather than .0
                    sub_children = child.children()
                    for sub_child in sub_children:
                        if sub_child.is_interface():
                            interface = sub_child.host
                            if interface.is_physical:

                                # physical interface

                                sub_child.ip_address = iterhosts.next()
                                sub_child.subnet = subnet
                            elif interface.is_loopback \
                                and not interface.is_loopback_zero:

                                # secondary loopback interface

                                sub_child.ip_address = iterhosts.next()
                                sub_child.subnet = subnet
                        else:
                            sub_child.subnet = iterhosts.next()

                        #log.debug('Allocate sub_child to %s %s'% (sub_child, sub_child.subnet))
                elif child.is_host():
                    child.subnet = subnet.next()
                elif child.is_loopback_group():
                    child.subnet = subnet.next()
                    iterhosts = child.subnet.iter_hosts()  # ensures start at .1 rather than .0
                    sub_children = child.children()
                    for sub_child in sub_children:
                        if sub_child.is_interface() \
                            and not sub_child.host.is_loopback_zero:

                           # secondary loopback

                            sub_child.ip_address = iterhosts.next()
                            sub_child.subnet = child.subnet
                        else:
                            sub_child.subnet = iterhosts.next()
                else:
                    child.subnet = subnet.next()
                    allocate(child)  # continue down the tree

        global_root.subnet = global_ip_block

# TODO: fix this workaround where referring to the wrong graph

        global_root_id = global_root.node
        global_root = TreeNode(global_graph, global_root_id)
        allocate(global_root)

# check for parentless nodes

        self.graph = global_graph
        self.root_node = global_root

    def group_allocations(self):
        allocs = {}
        for node in self:
            if node.group_attr:

                # TODO: Also need to store the type

                allocs[node.group_attr] = [node.subnet]

        return allocs

    def add_nodes(self, nodes):
        self.unallocated_nodes += list(nodes)

    def walk(self):

        def list_successors(node):
            successors = self.graph.successors(node)
            if successors:
                children = [list_successors(n) for n in successors]
                return {node: children}
            return node

        return list_successors(self.root_node)

    def json(self):

        def list_successors(node):
            children = node.children()
            if children:
                children = [list_successors(n) for n in children]
                return {'name': node, 'subnet': node.subnet,
                        'children': children}
            return {'name': node, 'subnet': node.subnet}

        if not self.root_node:
            log.debug('No root node set')
            return {'name': str(self.root_ip_block),
                    'subnet': str(self.root_ip_block), 'children': []}
            return
        return list_successors(self.root_node)

    def assign(self):

# assigns allocated addresses back to hosts
        # don't look at host nodes now - use loopback_groups
        #TODO: make check for interface and loopback zero now
        host_tree_nodes = [n for n in self if n.is_host()
                            and isinstance(n.host, autonetkit.anm.OverlayNode)
                           and n.host.is_l3device()]

        # for host_tree_node in host_tree_nodes:
            # print host_tree_node, host_tree_node.subnet

        for host_tree_node in host_tree_nodes:
            host_tree_node.host.loopback = host_tree_node.subnet

        cds = [n for n in self if n.is_broadcast_domain()]
        for cd in cds:
            cd.host.subnet = cd.subnet

        interfaces = [n for n in self if n.is_interface()]
        for n in interfaces:
            interface = n.host
            if interface.is_loopback and interface.is_loopback_zero:

                # primary loopback

                interface.loopback = n.ip_address
            elif interface.is_loopback \
                and not interface.is_loopback_zero:

                # secondary loopback

                interface.loopback = n.ip_address
                interface.subnet = n.subnet
            elif interface.is_physical:
                interface.ip_address = n.ip_address
                interface.subnet = n.subnet


def assign_asn_to_interasn_cds(g_ip, address_block=None):
    G_phy = g_ip.overlay('phy')
    for broadcast_domain in g_ip.nodes('broadcast_domain'):
        neigh_asn = list(ank_utils.neigh_attr(g_ip, broadcast_domain,
                         'asn', G_phy))  # asn of neighbors
        if len(set(neigh_asn)) == 1:
            asn = set(neigh_asn).pop()  # asn of any neigh, as all same
        else:
            asn = ank_utils.most_frequent(neigh_asn)  # allocate cd to asn with most neighbors in it
        broadcast_domain.asn = asn

    return


def allocate_infra(g_ip, address_block=None):
    if not address_block:
        address_block = netaddr.IPNetwork('10.0.0.0/8')
    log.info('Allocating v4 Infrastructure IPs')
    ip_tree = IpTree(address_block)
    assign_asn_to_interasn_cds(g_ip)
    ip_tree.add_nodes(g_ip.nodes('broadcast_domain'))
    ip_tree.build()

    # cd_tree = ip_tree.json()

    ip_tree.assign()

    g_ip.data.infra_blocks = ip_tree.group_allocations()

    # total_tree = { 'name': "ip", 'children': [cd_tree], }
    # jsontree = json.dumps(total_tree, cls=autonetkit.ank_json.AnkEncoder, indent = 4)

    g_ip.data.infra_blocks = ip_tree.group_allocations()


# TODO: apply directly here

def allocate_loopbacks(g_ip, address_block=None):
    if not address_block:
        address_block = netaddr.IPNetwork('192.168.0.0/22')
    log.info('Allocating v4 Primary Host loopback IPs')
    ip_tree = IpTree(address_block)
    ip_tree.add_nodes(g_ip.l3devices())
    ip_tree.build()

    # loopback_tree = ip_tree.json()

    ip_tree.assign()
    g_ip.data.loopback_blocks = ip_tree.group_allocations()


def allocate_vrf_loopbacks(g_ip, address_block=None):
    if not address_block:
        address_block = netaddr.IPNetwork('172.16.0.0/24')

    secondary_loopbacks = [i for n in g_ip.nodes() for i in
                           n.loopback_interfaces
                           if not i.is_loopback_zero]

    if not len(secondary_loopbacks):
        return   # nothing to set
    log.info('Allocating v4 Secondary Host loopback IPs')
    ip_tree = IpTree(address_block)

    vrf_loopbacks = [i for i in secondary_loopbacks if i['vrf'
                     ].vrf_name]

    ip_tree.add_nodes(vrf_loopbacks)

    ip_tree.build()

    # secondary_loopback_tree = ip_tree.json()

    ip_tree.assign()


    # TODO: store vrf block to g_ip.data



import autonetkit.ank as ank_utils
import autonetkit.log as log


from autonetkit.ank_utils import call_log

@call_log
def build_ibgp_v4(anm):
    #TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    #TODO: base on generic ibgp graph, rather than bgp graph
    g_bgp = anm['bgp']
    g_phy = anm['phy']
    g_ibgpv4 = anm.add_overlay("ibgp_v4", directed=True)
    ipv4_nodes = set(g_phy.routers("use_ipv4"))
    g_ibgpv4.add_nodes_from((n for n in g_bgp if n in ipv4_nodes),
            retain = ["ibgp_role", "hrr_cluster", "rr_cluster"] )
    g_ibgpv4.add_edges_from(g_bgp.edges(type="ibgp"), retain="direction")

@call_log
def build_ibgp_v6(anm):
    #TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    #TODO: base on generic ibgp graph, rather than bgp graph
    g_bgp = anm['bgp']
    g_phy = anm['phy']
    g_ibgpv6 = anm.add_overlay("ibgp_v6", directed=True)
    ipv6_nodes = set(g_phy.routers("use_ipv6"))
    g_ibgpv6.add_nodes_from((n for n in g_bgp if n in ipv6_nodes),
            retain = ["ibgp_role", "hrr_cluster", "rr_cluster"] )
    g_ibgpv6.add_edges_from(g_bgp.edges(type="ibgp"), retain="direction")

@call_log
def build_ebgp_v4(anm):
    #TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    g_ebgp = anm['ebgp']
    g_phy = anm['phy']
    g_ebgpv4 = anm.add_overlay("ebgp_v4", directed=True)
    ipv4_nodes = set(g_phy.routers("use_ipv4"))
    g_ebgpv4.add_nodes_from(n for n in g_ebgp if n in ipv4_nodes)
    g_ebgpv4.add_edges_from(g_ebgp.edges(), retain="direction")

def build_ebgp_v6(anm):
    #TODO: remove the bgp layer and have just ibgp and ebgp
    # TODO: build from design rules, currently just builds from ibgp links in bgp layer
    g_ebgp = anm['ebgp']
    g_phy = anm['phy']
    g_ebgpv6 = anm.add_overlay("ebgp_v6", directed=True)
    ipv6_nodes = set(g_phy.routers("use_ipv6"))
    g_ebgpv6.add_nodes_from(n for n in g_ebgp if n in ipv6_nodes)
    g_ebgpv6.add_edges_from(g_ebgp.edges(), retain="direction")


@call_log
def build_ebgp(anm):
    g_in = anm['input']
    g_phy = anm['phy']
    g_ebgp = anm.add_overlay("ebgp", directed=True)
    g_ebgp.add_nodes_from(g_in.routers())
    ebgp_edges = [e for e in g_in.edges() if not e.attr_equal("asn")]
    g_ebgp.add_edges_from(ebgp_edges, bidirectional=True, type='ebgp')

    ebgp_switches = [n for n in g_in.switches()
            if not ank_utils.neigh_equal(g_phy, n, "asn")]
    g_ebgp.add_nodes_from(ebgp_switches, retain=['asn'])
    g_ebgp.log.debug("eBGP switches are %s" % ebgp_switches)
    g_ebgp.add_edges_from((e for e in g_in.edges()
            if e.src in ebgp_switches or e.dst in ebgp_switches),
    bidirectional=True, type='ebgp')
    ank_utils.aggregate_nodes(g_ebgp, ebgp_switches)
    # need to recalculate as may have aggregated
    ebgp_switches = list(g_ebgp.switches())
    g_ebgp.log.debug("aggregated eBGP switches are %s" % ebgp_switches)
    exploded_edges = ank_utils.explode_nodes(g_ebgp, ebgp_switches)
    same_asn_edges = []
    for edge in exploded_edges:
        if edge.src.asn == edge.dst.asn:
            same_asn_edges.append(edge)
        else:
            edge.multipoint = True
    """TODO: remove up to here once compiler updated"""

    g_ebgp.remove_edges_from(same_asn_edges)


@call_log
def build_ibgp(anm):
    g_in = anm['input']
    g_phy = anm['phy']
    g_bgp = anm['bgp']

    #TODO: build direct to ibgp graph - can construct combined bgp for vis

    ank_utils.copy_attr_from(g_in, g_bgp, "ibgp_role")
    ank_utils.copy_attr_from(g_in, g_bgp, "ibgp_l2_cluster", "hrr_cluster", default = None)
    ank_utils.copy_attr_from(g_in, g_bgp, "ibgp_l3_cluster", "rr_cluster", default = None)

    #TODO: add more detailed logging

    for n in g_bgp:
        # Tag with label to make logic clearer
        if n.ibgp_role is None:
            n.ibgp_role = "Peer"

            #TODO: if top-level, then don't mark as RRC

    ibgp_nodes = [n for n in g_bgp if not n.ibgp_role is "Disabled"]

    # Notify user of non-ibgp nodes
    non_ibgp_nodes = [n for n in g_bgp if n.ibgp_role is "Disabled"]
    if len(non_ibgp_nodes) < 10:
        log.info("Skipping iBGP for iBGP disabled nodes: %s" % non_ibgp_nodes)
    elif len(non_ibgp_nodes) >= 10:
        log.info("Skipping iBGP for more than  10 iBGP disabled nodes: refer to visualization for resulting topology.")

    for asn, asn_devices in ank_utils.groupby("asn", ibgp_nodes):
        asn_devices = list(asn_devices)

        # iBGP peer peers with
        peers = [n for n in asn_devices if n.ibgp_role == "Peer"]
        rrs = [n for n in asn_devices if n.ibgp_role == "RR"]
        hrrs = [n for n in asn_devices if n.ibgp_role == "HRR"]
        rrcs = [n for n in asn_devices if n.ibgp_role == "RRC"]


        over_links = []
        up_links = []
        down_links = []

        # 0. RRCs can only belong to either an rr_cluster or a hrr_cluster
        invalid_rrcs = [r for r in rrcs if r.rr_cluster is not None and r.hrr_cluster is not None]
        if len(invalid_rrcs):
            message = ", ".join(str(r) for r in invalid_rrcs)
            log.warning("RRCs can only have either a rr_cluster or hrr_cluster set. "
                "The following have both set, and only the rr_cluster will be used: %s" % message)

        #TODO: do we also want to warn for RRCs with no cluster set? Do we also exclude these?
        #TODO: do we also want to warn for HRRs and RRs with no cluster set? Do we also exclude these?

        # 1. Peers:
        # 1a. Peers connect over to peers
        over_links += [(s,t) for s in peers for t in peers]
        # 1b. Peers connect over to RRs
        over_links += [(s,t) for s in peers for t in rrs]

        # 2. RRs:
        # 2a. RRs connect over to Peers
        over_links += [(s,t) for s in rrs for t in peers]
        # 2b. RRs connect over to RRs
        over_links += [(s,t) for s in rrs for t in rrs]
        # 2c. RRs connect down to RRCs in same rr_cluster
        down_links += [(s,t) for s in rrs for t in rrcs
            if s.rr_cluster == t.rr_cluster != None]
        # 2d. RRs connect down to HRRs in the same rr_cluster
        down_links += [(s,t) for s in rrs for t in hrrs
            if s.rr_cluster == t.rr_cluster != None]

        # 3. HRRs
        # 3a. HRRs connect up to RRs in the same rr_cluster
        up_links += [(s,t) for s in hrrs for t in rrs
            if s.rr_cluster == t.rr_cluster != None]
        # 3b. HRRs connect down to RRCs in same hrr_cluster (providing RRC has no rr_cluster set)
        down_links += [(s,t) for s in hrrs for t in rrcs
            if s.hrr_cluster == t.hrr_cluster != None
            and t.rr_cluster is None]

        # 4. RRCs
        # 4a. RRCs connect up to RRs in the same rr_cluster (regardless if RRC has hrr_cluster set)
        up_links += [(s,t) for s in rrcs for t in rrs
            if s.rr_cluster == t.rr_cluster != None]
        # 3b. RRCs connect up to HRRs in same hrr_cluster (providing RRC has no rr_cluster set)
        up_links += [(s,t) for s in rrcs for t in hrrs
            if s.hrr_cluster == t.hrr_cluster != None
            and s.rr_cluster is None]

        # Remove self-links
        over_links = [(s,t) for s,t in over_links if s!=t]
        up_links = [(s,t) for s,t in up_links if s!=t]
        down_links = [(s,t) for s,t in down_links if s!=t]

        g_bgp.add_edges_from(over_links, type='ibgp', direction='over')
        g_bgp.add_edges_from(up_links, type='ibgp', direction='up')
        g_bgp.add_edges_from(down_links, type='ibgp', direction='down')

@call_log
def build_bgp(anm):
    """Build iBGP end eBGP overlays"""
    # eBGP
    g_in = anm['input']
    g_phy = anm['phy']

    if not anm['phy'].data.enable_routing:
        log.info("Routing disabled, not configuring BGP")
        return

    build_ebgp(anm)
    build_ebgp_v4(anm)
    build_ebgp_v6(anm)

    """TODO: remove from here once compiler updated"""
    g_bgp = anm.add_overlay("bgp", directed=True)
    g_bgp.add_nodes_from(g_in.routers())
    ebgp_edges = [edge for edge in g_in.edges() if not edge.attr_equal("asn")]
    g_bgp.add_edges_from(ebgp_edges, bidirectional=True, type='ebgp')

    ebgp_switches = [n for n in g_in.switches()
            if not ank_utils.neigh_equal(g_phy, n, "asn")]
    g_bgp.add_nodes_from(ebgp_switches, retain=['asn'])
    log.debug("eBGP switches are %s" % ebgp_switches)
    g_bgp.add_edges_from((e for e in g_in.edges()
            if e.src in ebgp_switches or e.dst in ebgp_switches), bidirectional=True, type='ebgp')
    ank_utils.aggregate_nodes(g_bgp, ebgp_switches)
    ebgp_switches = list(g_bgp.switches()) # need to recalculate as may have aggregated
    log.debug("aggregated eBGP switches are %s" % ebgp_switches)
    exploded_edges = ank_utils.explode_nodes(g_bgp, ebgp_switches)

    same_asn_edges = []
    for edge in exploded_edges:
        if edge.src.asn == edge.dst.asn:
            same_asn_edges.append(edge)
        else:
            edge.multipoint = True
    """TODO: remove up to here once compiler updated"""

    g_bgp.remove_edges_from(same_asn_edges)


    build_ibgp(anm)

    ebgp_nodes = [d for d in g_bgp if any(
        edge.type == 'ebgp' for edge in d.edges())]
    g_bgp.update(ebgp_nodes, ebgp=True)

    for ebgp_edge in g_bgp.edges(type = "ebgp"):
        for interface in ebgp_edge.interfaces():
            interface.ebgp = True

    for edge in g_bgp.edges(type='ibgp'):
        # TODO: need interface querying/selection. rather than hard-coded ids
        edge.bind_interface(edge.src, 0)

    #TODO: need to initialise interface zero to be a loopback rather than physical type
    for node in g_bgp:
        for interface in node.interfaces():
            interface.multipoint = any(e.multipoint for e in interface.edges())

    build_ibgp_v4(anm)
    build_ibgp_v6(anm)



#!/usr/bin/python
# -*- coding: utf-8 -*-
import socket

import autonetkit.ank_json
import autonetkit.config as config
import autonetkit.log as log
from autonetkit.ank_utils import call_log

use_http_post = config.settings['Http Post']['active']
if use_http_post:
    import urllib


@call_log
def format_http_url(host=None, port=None, route='publish'):
    if not host and not port:
        host = config.settings['Http Post']['server']
        port = config.settings['Http Post']['port']
    return 'http://%s:%s/%s' % (host, port, route)


default_http_url = format_http_url()

@call_log
def update_http(
    anm=None,
    nidb=None,
    http_url=None,
    uuid=None,
    ):
    if http_url is None:
        http_url = default_http_url

    if anm and nidb:
        body = autonetkit.ank_json.dumps(anm, nidb)
    elif anm:
        body = autonetkit.ank_json.dumps(anm)
    else:
        import json
        body = json.dumps({})  # blank to test visualisation server running

    if uuid is None:
        uuid = get_uuid(anm)

    params = urllib.urlencode({'body': body, 'type': 'anm',
                              'uuid': uuid})
    try:
        data = urllib.urlopen(http_url, params).read()
        log.debug(data)
    except IOError, e:
        log.info('Unable to connect to visualisation server %s'
                 % http_url)
        return

    if not anm:

        # testing

        log.info('Visualisation server running')

@call_log
def get_uuid(anm):
    try:
        return config.settings['Http Post']['uuid']
    except KeyError:
        log.warning('UUID not set, returning singleuser uuid')
        return 'singleuser'


@call_log
def highlight(
    nodes=None,
    edges=None,
    paths=None,
    path=None,
    uuid='singleuser',
    http_url=None,
    ):
    if http_url is None:
        http_url = default_http_url
    if not paths:
        paths = []

    if path:
        paths.append(path)

    if nodes is None:
        nodes = []
    if edges is None:
        edges = []

    def nfilter(n):
        try:
            return n.id
        except AttributeError:
            return n  # likely already a node id (string)

    def efilter(e):
        try:
            return (e.src.id, e.dst.id)
        except AttributeError:
            return e  # likely already edge (src, dst) id tuple (string)

    nodes = [nfilter(n) for n in nodes]
    #TODO: allow node annotations also

    filtered_edges = []
    for edge in edges:
        if isinstance(edge, dict) and 'edge' in edge:
           edge_data = dict(edge)  # use as-is (but make copy)
        else:
            edge_data = {'edge': edge} # no extra data

        edge_data['src'] = edge_data['edge'].src.id
        edge_data['dst'] = edge_data['edge'].dst.id
        del edge_data['edge'] # remove now have extracted the src/dst
        filtered_edges.append(edge_data)

    #edges = [efilter(e) for e in edges]
    filtered_paths = []
    for path in paths:

        # TODO: tidy this logic

        if isinstance(path, dict) and 'path' in path:
            path_data = dict(path ) # use as-is (but make copy)
        else:
            import random
            is_verified = bool(random.randint(0, 1))

            # path_data = {'path': path, 'verified': is_verified}

            path_data = {'path': path}

        path_data['path'] = [nfilter(n) for n in path_data['path']]
        filtered_paths.append(path_data)

    # TODO: remove "highlight" from json, use as url params to distinguish

    import json
    body = json.dumps({'nodes': nodes, 'edges': filtered_edges,
                      'paths': filtered_paths})

    params = urllib.urlencode({'body': body, 'type': 'highlight',
                              'uuid': uuid})

    # TODO: split this common function out, create at runtime so don't need to keep reading config

    try:
        data = urllib.urlopen(http_url, params).read()
    except IOError, e:
        log.info('Unable to connect to HTTP Server %s: %s' % (http_url,
                 e))

